#include "stdafx.h"
#include "Tif2Bmp.h"

bool IsBlackWhite(CVisByteImage &cfaImage)
{
	double diff = 0;
	double avg = 0;
	for (int x = 0; x < cfaImage.Width(); x++)
		for (int y = 0; y < cfaImage.Height(); y++)
		{
			avg += cfaImage.Pixel(x,y);
			if (x-1>=0)
				diff += abs(cfaImage.Pixel(x,y)-cfaImage.Pixel(x-1,y));
			if (y-1>=0)
				diff += abs(cfaImage.Pixel(x,y)-cfaImage.Pixel(x,y-1));
			if (x+1<cfaImage.Width())
				diff += abs(cfaImage.Pixel(x,y)-cfaImage.Pixel(x+1,y));
			if (y+1<cfaImage.Height())
				diff += abs(cfaImage.Pixel(x,y)-cfaImage.Pixel(x,y+1));
		}
	return (diff < avg/2.5);
}

void FreeSpline(int n, double y[])
{
	int i;
	double *A = new double[n+1];
	double *u = new double[n+1];
	double *z = new double[n+1];
	double *M = new double[n+1];

	A[0] = 0;
	for (i = 1; i < n; i++)
		A[i] = (y[2*(i+1)]-y[2*i])-(y[2*i]-y[2*(i-1)]);
	A[n] = 0;

	z[0] = 0;
	u[0] = 0;
	for (i = 1; i <= n; i++)
	{
		z[i] = (A[i]-z[i-1]/6.0)/(2.0/3.0-u[i-1]/6.0);
		u[i] = 1.0/6.0/(2.0/3.0-u[i-1]/6.0);
	}
	M[n] = z[n];
	for (i = n-1; i >= 0; i--)
		M[i] = z[i]-u[i]*M[i+1];

	for (i = 0; i < n; i++)
	{
		double C = (y[2*(i+1)]-y[2*i])-(M[i+1]-M[i])/6.0;
		double D = y[2*i]-M[i]/6.0-C*i;

		y[2*i+1] = (M[i]+M[i+1])/48.0+C*(i+0.5)+D;
		if (y[2*i+1] > 255)
			y[2*i+1] = 255;
		else if (y[2*i+1] < 0)
			y[2*i+1] = 0;
	}
}

void ToColor(CVisByteImage &cfa, CVisRGBAByteImage &rgb)
{
	rgb.Allocate(cfa.Width(),cfa.Height());

	int dx = cfa.Width();
	int dy = cfa.Height();

	int x,y;
	int hdiff,vdiff;

	// green interpolation
	for (x = 2; x < dx-2; x++)
		for (y = 2; y < dy-2; y++)
			if ((x%2)==(y%2))
			{
				hdiff=abs(((cfa.Pixel(x-2,y)+cfa.Pixel(x+2,y))/2)-cfa.Pixel(x,y));
				vdiff=abs(((cfa.Pixel(x,y-2)+cfa.Pixel(x,y+2))/2)-cfa.Pixel(x,y));
				if (hdiff<vdiff)
					rgb.Pixel(x,y).SetG((cfa.Pixel(x-1,y)+cfa.Pixel(x+1,y))/2);
				else if (vdiff<hdiff)
					rgb.Pixel(x,y).SetG((cfa.Pixel(x,y-1)+cfa.Pixel(x,y+1))/2);
				else
					rgb.Pixel(x,y).SetG((cfa.Pixel(x-1,y)+cfa.Pixel(x+1,y)+cfa.Pixel(x,y-1)+cfa.Pixel(x,y+1))/4);
			}
			else
				rgb.Pixel(x,y).SetG(cfa.Pixel(x,y));

	// red interpolation
	for (x = 2; x < dx-2; x += 2)
		for (y = 2; y < dy-2; y += 2)
		{
			rgb.Pixel(x  ,y  ).SetR( cfa.Pixel(x  ,y  ));
			rgb.Pixel(x+1,y  ).SetR(min(255,max(0,
															(cfa.Pixel(x  ,y  )-rgb.Pixel(x  ,y  ).G()
															+cfa.Pixel(x+2,y  )-rgb.Pixel(x+2,y  ).G())/2
															+rgb.Pixel(x+1,y  ).G())));
			rgb.Pixel(x  ,y+1).SetR(min(255,max(0,
															(cfa.Pixel(x  ,y  )-rgb.Pixel(x  ,y  ).G()
															+cfa.Pixel(x  ,y+2)-rgb.Pixel(x  ,y+2).G())/2
															+rgb.Pixel(x  ,y+1).G())));
			rgb.Pixel(x+1,y+1).SetR(min(255,max(0,
															(cfa.Pixel(x  ,y  )-rgb.Pixel(x  ,y  ).G()
															+cfa.Pixel(x+2,y  )-rgb.Pixel(x+2,y  ).G()
															+cfa.Pixel(x  ,y+2)-rgb.Pixel(x  ,y+2).G()
															+cfa.Pixel(x+2,y+2)-rgb.Pixel(x+2,y+2).G())/4
															+rgb.Pixel(x+1,y+1).G())));
		}

	// blue interpolation
	for (x = 3; x < dx-1; x += 2)
		for (y = 3; y < dy-1; y += 2)
		{
			rgb.Pixel(x  ,y  ).SetB( cfa.Pixel(x  ,y  ));
			rgb.Pixel(x+1,y  ).SetB(min(255,max(0,
															(cfa.Pixel(x  ,y  )-rgb.Pixel(x  ,y  ).G()
															+cfa.Pixel(x+2,y  )-rgb.Pixel(x+2,y  ).G())/2
															+rgb.Pixel(x+1,y  ).G())));
			rgb.Pixel(x  ,y+1).SetB(min(255,max(0,
															(cfa.Pixel(x  ,y  )-rgb.Pixel(x  ,y  ).G()
															+cfa.Pixel(x  ,y+2)-rgb.Pixel(x  ,y+2).G())/2
															+rgb.Pixel(x  ,y+1).G())));
			rgb.Pixel(x+1,y+1).SetB(min(255,max(0,
															(cfa.Pixel(x  ,y  )-rgb.Pixel(x  ,y  ).G()
															+cfa.Pixel(x+2,y  )-rgb.Pixel(x+2,y  ).G()
															+cfa.Pixel(x  ,y+2)-rgb.Pixel(x  ,y+2).G()
															+cfa.Pixel(x+2,y+2)-rgb.Pixel(x+2,y+2).G())/4
															+rgb.Pixel(x+1,y+1).G())));
		}

	for (x = 3; x < dx-1; x++)
		rgb.Pixel(x,dy-2).SetB(0);
	for (y = 3; y < dy-1; y++)
		rgb.Pixel(dx-2,y).SetB(0);
}

void InterpolateBMP(CVisByteImage &cfa, CVisRGBAByteImage &rgb)
{
	rgb.Allocate(cfa.Width(),cfa.Height());

	double P[644];
	int row, col;

	// Interpolate Red
	for (row = 0; row < cfa.Height(); row += 2)
	{
		for (col = 0; col < cfa.Width(); col += 2)
			P[col] = cfa.Pixel(col,row);
		FreeSpline(cfa.Width()/2,P);
		for (col = 0; col < cfa.Width()-1; col ++)
			rgb.Pixel(col,row).SetR(P[col]);
		rgb.Pixel(col,row).SetR(P[col-2]);
	}
	for (col = 0; col < cfa.Width(); col ++)
	{
		for (row = 0; row < cfa.Height(); row += 2)
			P[row] = rgb.Pixel(col,row).R();
		FreeSpline(cfa.Height()/2,P);
		for (row = 1; row < cfa.Height()-1; row += 2)
			rgb.Pixel(col,row).SetR(P[row]);
		rgb.Pixel(col,row).SetR(P[row-2]);
	}

	// Interpolate Green
	for (row = 0; row < cfa.Height(); row += 2)
	{
		for (col = 1; col < cfa.Width(); col += 2)
			P[col-1] = cfa.Pixel(col,row);
		FreeSpline(cfa.Width()/2,P);
		rgb.Pixel(0,row).SetG(P[0]);
		for (col = 1; col < cfa.Width(); col ++)
			rgb.Pixel(col,row).SetG(P[col-1]);
	}
	for (row = 1; row < cfa.Height(); row += 2)
	{
		for (col = 0; col < cfa.Width(); col += 2)
			P[col] = cfa.Pixel(col,row);
		FreeSpline(cfa.Width()/2,P);
		for (col = 0; col < cfa.Width()-1; col ++)
			rgb.Pixel(col,row).SetG(P[col]);
		rgb.Pixel(col,row).SetG(P[col-2]);
	}

	// Interpolate Blue
	for (row = 1; row < cfa.Height(); row += 2)
	{
		for (col = 1; col < cfa.Width(); col += 2)
			P[col-1] = cfa.Pixel(col,row);
		FreeSpline(cfa.Width()/2,P);
		rgb.Pixel(0,row).SetB(P[0]);
		for (col = 1; col < cfa.Width(); col ++)
			rgb.Pixel(col,row).SetB(P[col-1]);
	}
	for (col = 0; col < cfa.Width(); col ++)
	{
		for (row = 1; row < cfa.Height(); row += 2)
			P[row-1] = rgb.Pixel(col,row).B();
		FreeSpline(cfa.Height()/2,P);
		rgb.Pixel(col,0).SetB(P[0]);
		for (row = 2; row < cfa.Height(); row += 2)
			rgb.Pixel(col,row).SetB(P[row-1]);
	}
}

void TIF2LoBMP(CVisByteImage &cfaImage, CVisRGBAByteImage &rgbImage)
{
	rgbImage.Allocate(cfaImage.Width()/2,cfaImage.Height()/2);

	for (int x = 0; x < rgbImage.Width(); x++)
		for (int y = 0; y < rgbImage.Height(); y++)
		{
			rgbImage.Pixel(x,y).SetR(cfaImage.Pixel(2*x,2*y));
			rgbImage.Pixel(x,y).SetG((cfaImage.Pixel(2*x,2*y+1)+cfaImage.Pixel(2*x+1,2*y))/2);
			rgbImage.Pixel(x,y).SetB(cfaImage.Pixel(2*x+1,2*y+1));
			rgbImage.Pixel(x,y).SetA(0);
		}
}

void TIF2HiBMP(CVisByteImage &cfaImage, CVisRGBAByteImage &rgbImage)
{
	rgbImage.Allocate(cfaImage.Width(),cfaImage.Height());

	for (int x = 0; x < rgbImage.Width(); x++)
		for (int y = 0; y < rgbImage.Height(); y++)
		{
			rgbImage.Pixel(x,y).SetA(0);

			if ((x%2 == 0) && (y%2 == 0))
			{
				rgbImage.Pixel(x,y).SetR(cfaImage.Pixel(x,y));
				rgbImage.Pixel(x,y).SetG(Average4(cfaImage,x,y));
				rgbImage.Pixel(x,y).SetB(Average3(cfaImage,x,y));
			}
			else if ((x%2 == 0) && (y%2 == 1))
			{
				rgbImage.Pixel(x,y).SetR(Average1(cfaImage,x,y));
				rgbImage.Pixel(x,y).SetG(Average5(cfaImage,x,y));
				rgbImage.Pixel(x,y).SetB(Average2(cfaImage,x,y));
			}
			else if ((x%2 == 1) && (y%2 == 0))
			{
				rgbImage.Pixel(x,y).SetR(Average2(cfaImage,x,y));
				rgbImage.Pixel(x,y).SetG(Average5(cfaImage,x,y));
				rgbImage.Pixel(x,y).SetB(Average1(cfaImage,x,y));
			}
			else //if ((x%2 == 1) && (y%2 == 1))
			{
				rgbImage.Pixel(x,y).SetR(Average3(cfaImage,x,y));
				rgbImage.Pixel(x,y).SetG(Average4(cfaImage,x,y));
				rgbImage.Pixel(x,y).SetB(cfaImage.Pixel(x,y));
			}
		}
}

void WhiteBalance(CVisRGBAByteImage &rgbImage)
{
	for (int x = 0; x < rgbImage.Width(); x++)
		for (int y = 0; y < rgbImage.Height(); y++)
		{
/*
			if (rgbImage.Pixel(x,y).R() < 255)
				rgbImage.Pixel(x,y).SetR(rgbImage.Pixel(x,y).R());
			else
				rgbImage.Pixel(x,y).SetR(255);
			if (rgbImage.Pixel(x,y).G() < 255)
				rgbImage.Pixel(x,y).SetG(rgbImage.Pixel(x,y).G());
			else
				rgbImage.Pixel(x,y).SetG(255);
			if (rgbImage.Pixel(x,y).B()*2 < 255)
				rgbImage.Pixel(x,y).SetB(rgbImage.Pixel(x,y).B()*2);
			else
				rgbImage.Pixel(x,y).SetB(255);
*/
			if (rgbImage.Pixel(x,y).R()*1.55 < 255)
				rgbImage.Pixel(x,y).SetR(rgbImage.Pixel(x,y).R()*1.55);
			else
				rgbImage.Pixel(x,y).SetR(255);
			if (rgbImage.Pixel(x,y).G()*1.4 < 255)
				rgbImage.Pixel(x,y).SetG(rgbImage.Pixel(x,y).G()*1.4);
			else
				rgbImage.Pixel(x,y).SetG(255);
			if (rgbImage.Pixel(x,y).B()*2.8 < 255)
				rgbImage.Pixel(x,y).SetB(rgbImage.Pixel(x,y).B()*2.8);
			else
				rgbImage.Pixel(x,y).SetB(255);
		}
}

bool IsValid(CVisByteImage &cfaImage, int x, int y)
{
	return ((x >= 0) && (y >= 0) && (x < cfaImage.Width()) && (y < cfaImage.Height()));
}

int Average1(CVisByteImage &cfaImage, int x, int y)
{
	int total = 0;
	int count = 0;

	if (IsValid(cfaImage,x,y-1)) { total += cfaImage.Pixel(x,y-1), count++; }
	if (IsValid(cfaImage,x,y+1)) { total += cfaImage.Pixel(x,y+1), count++; }
	return total/count;
}

int Average2(CVisByteImage &cfaImage, int x, int y)
{
	int total = 0;
	int count = 0;

	if (IsValid(cfaImage,x-1,y)) { total += cfaImage.Pixel(x-1,y), count++; }
	if (IsValid(cfaImage,x+1,y)) { total += cfaImage.Pixel(x+1,y), count++; }
	return total/count;
}

int Average3(CVisByteImage &cfaImage, int x, int y)
{
	int total = 0;
	int count = 0;

	if (IsValid(cfaImage,x-1,y-1)) { total += cfaImage.Pixel(x-1,y-1), count++; }
	if (IsValid(cfaImage,x-1,y+1)) { total += cfaImage.Pixel(x-1,y+1), count++; }
	if (IsValid(cfaImage,x+1,y-1)) { total += cfaImage.Pixel(x+1,y-1), count++; }
	if (IsValid(cfaImage,x+1,y+1)) { total += cfaImage.Pixel(x+1,y+1), count++; }
	return total/count;
}

int Average4(CVisByteImage &cfaImage, int x, int y)
{
	int total = 0;
	int count = 0;

	if (IsValid(cfaImage,x-1,y)) { total += cfaImage.Pixel(x-1,y), count++; }
	if (IsValid(cfaImage,x+1,y)) { total += cfaImage.Pixel(x+1,y), count++; }
	if (IsValid(cfaImage,x,y-1)) { total += cfaImage.Pixel(x,y-1), count++; }
	if (IsValid(cfaImage,x,y+1)) { total += cfaImage.Pixel(x,y+1), count++; }
	return total/count;
}

int Average5(CVisByteImage &cfaImage, int x, int y)
{
	int total = 0;
	int count = 0;

	total += 4*cfaImage.Pixel(x,y), count+=4;

	if (IsValid(cfaImage,x-1,y-1)) { total += cfaImage.Pixel(x-1,y-1), count++; }
	if (IsValid(cfaImage,x-1,y+1)) { total += cfaImage.Pixel(x-1,y+1), count++; }
	if (IsValid(cfaImage,x+1,y-1)) { total += cfaImage.Pixel(x+1,y-1), count++; }
	if (IsValid(cfaImage,x+1,y+1)) { total += cfaImage.Pixel(x+1,y+1), count++; }
	return total/count;
}
